""""""
import traitlets
from nbconvert import preprocessors

class LiteraryTagAllowListPreprocessor(preprocessors.Preprocessor):
    allow_cell_tags = traitlets.Set(traitlets.Unicode(), default_value={'export', 'docstring'})

    def check_cell_conditions(self, cell, resources: dict, index: int) -> bool:
        tags = cell.metadata.get('tags', [])
        return bool(self.allow_cell_tags.intersection(tags))

    def preprocess(self, nb, resources: dict):
        nb.cells = [self.preprocess_cell(cell, resources, i)[0] for (i, cell) in enumerate(nb.cells) if self.check_cell_conditions(cell, resources, i)]
        return (nb, resources)

    def preprocess_cell(self, cell, resources: dict, index: int):
        return (cell, resources)


"""# Testing
"""
import logging
import pathlib
import nbclient
import nbformat
logger = logging.getLogger(__name__)

def run_notebook(path: pathlib.Path, cell_timeout: int=20):
    """Run a Jupyter notebook at the given path

    :param path: path to notebook
    :param cell_timeout: timeout for each cell execution
    :return:
    """
    try:
        logger.info(f'Executing {path}')
        nb = nbformat.read(path, as_version=nbformat.NO_CONVERT)
        client = nbclient.NotebookClient(nb, timeout=cell_timeout, resources={'metadata': {'path': path.parent}})
        client.execute()
        logger.info(f'Finished executing {path}')
    except Exception:
        logger.exception(f'Error during execution of {path}')
        raise


"""# Test Package Command
"""
import logging
import pathlib
from concurrent import futures
from jupyter_core.application import JupyterApp
from traitlets import Int, List, Unicode, default
from ..core import testing
from ..core.config import CONFIG_FILE_NAME
logger = logging.getLogger(__name__)
DEFAULT_IGNORE_PATTERNS = ('.ipynb_checkpoints',)

def _patch_nbclient_exceptions():
    """Shim to patch nbclient exception pickling"""
    import nbclient.exceptions as exc

    def reduce_CellExecutionError(self):
        return (type(self), (self.traceback, self.ename, self.evalue))
    exc.CellExecutionError.__reduce__ = reduce_CellExecutionError

def find_notebooks(path, ignore_patterns=None):
    """Find notebooks given by a particular path.

    If the path is a directory, yield from the result of calling find_notebooks` with
    the directory path.
    If the path is a notebook file path, yield the path directly

    :param path: path to a file or directory
    :param ignore_patterns: set of patterns to ignore during recursion
    :return:
    """
    if ignore_patterns is None:
        ignore_patterns = DEFAULT_IGNORE_PATTERNS
    if path.is_dir():
        for p in path.iterdir():
            if any((path.match(p) for p in ignore_patterns)):
                continue
            yield from find_notebooks(p, ignore_patterns)
    elif path.match('*.ipynb'):
        yield path

class LiteraryTestApp(JupyterApp):
    name = 'literary test'
    description = 'Test literary notebooks in parallel'
    aliases = {**JupyterApp.aliases, 'source': 'LiteraryTestApp.source', 'jobs': 'LiteraryTestApp.jobs', 'ignore': 'LiteraryTestApp.ignore'}
    source = List(trait=Unicode(help='source directory or notebooks to run')).tag(config=True)
    jobs = Int(allow_none=True, default_value=None, help='number of parallel jobs to run').tag(config=True)
    ignore = List(help='glob pattern to ignore during recursion', trait=Unicode()).tag(config=True)

    @default('config_file_name')
    def _config_file_name_default(self):
        return CONFIG_FILE_NAME

    def start(self):
        if not self.source:
            raise ValueError(f'Missing source path(s)')
        source = [pathlib.Path(s) for s in self.source]
        paths = [p.resolve() for s in source for p in find_notebooks(s, self.ignore)]
        with futures.ProcessPoolExecutor(max_workers=self.jobs, initializer=_patch_nbclient_exceptions) as executor:
            tasks = [executor.submit(testing.run_notebook, p) for p in paths]
            for task in futures.as_completed(tasks):
                task.result()


"""# Build Package Command
"""
import pathlib
from jupyter_core.application import JupyterApp
from traitlets import default, Unicode, List
from ..core.config import CONFIG_FILE_NAME
from ..core.package import build_package

class LiteraryBuildApp(JupyterApp):
    name = 'literary build'
    description = 'Build a pure-Python package from a set of Jupyter notebooks'
    aliases = {**JupyterApp.aliases, 'source': 'LiteraryBuildApp.source', 'package': 'LiteraryBuildApp.package', 'ignore': 'LiteraryBuildApp.ignore'}
    source = Unicode(help='source directory for notebooks').tag(config=True)
    package = Unicode(help='destination path generated package').tag(config=True)
    ignore = List(help='glob pattern to ignore during recursion', trait=Unicode()).tag(config=True)

    @default('config_file_name')
    def _config_file_name_default(self):
        return CONFIG_FILE_NAME

    def start(self):
        if not self.source:
            raise ValueError(f'Missing source path')
        source = pathlib.Path(self.source)
        if not self.package:
            raise ValueError(f'Missing package path')
        package = pathlib.Path(self.package)
        build_package(source, package, self.ignore)


"""# Exporter
"""
import ast
import logging
import sys
import traitlets.config
from nbconvert import exporters
from traitlets import List, default, import_item
from .preprocessor import LiteraryTagAllowListPreprocessor
from .utils import escape_triple_quotes, unparse_ast
logger = logging.getLogger(__name__)

class LiteraryPythonExporter(exporters.PythonExporter):

    def __init__(self, *args, **kwargs):
        """
        Public constructor

        Parameters
        ----------
        config : ``traitlets.config.Config``
            User configuration instance.
        `**kw`
            Additional keyword arguments passed to parent __init__

        """
        super().__init__(*args, **kwargs)
        self._init_transformers()
    transformers = List(default_value=['literary.core.transformers.PatchTransformer', 'literary.core.transformers.IPythonTransformer']).tag(config=True)

    def _init_transformers(self):
        self._transformers = []
        for value in self.transformers:
            if isinstance(value, str):
                value = import_item(value)
            self._transformers.append(value())

    @default('template_name')
    def _template_name_default(self):
        return 'literary'

    @default('default_preprocessors')
    def _default_preprocessors_default(self):
        return [LiteraryTagAllowListPreprocessor]

    @default('exclude_input_prompt')
    def _exclude_input_prompt_default(self):
        return True

    @property
    def default_config(self):
        c = traitlets.config.Config({'LiteraryTagAllowListPreprocessor': {'enabled': True}})
        c.merge(super().default_config)
        return c

    def default_filters(self):
        yield from super().default_filters()
        yield ('escape_triple_quotes', escape_triple_quotes)

    def from_notebook_node(self, nb, resources=None, **kwargs):
        (body, resources) = super().from_notebook_node(nb, resources, **kwargs)
        node = ast.parse(body)
        try:
            for transformer in self._transformers:
                node = transformer.visit(node)
        except Exception as err:
            raise RuntimeError(f'An error occurred during AST transforming: {body}') from err
        return (unparse_ast(node), resources)


"""# Utility Functions
"""
import sys
from typing_extensions import Literal

def escape_triple_quotes(string: str, single_quote: Literal["'", '"']='"') -> str:
    """Escape triple quotes inside a string

    :param string: string to escape
    :param single_quote: single-quote character
    :return: escaped string
    """
    assert len(single_quote) == 1
    quote = single_quote * 3
    escaped_single_quote = f'\\{single_quote}'
    escaped_quote = escaped_single_quote * 3
    return string.replace(quote, escaped_quote)
if sys.version_info < (3, 9, 0):
    import astunparse
    import astunparse.unparser

    class ASTUnparser(astunparse.unparser.Unparser):
        """AST unparser with additional preference for triple-quoted multi-line strings"""

        def _Constant(self, tree):
            if isinstance(tree.value, str) and '\n' in tree.value:
                self.write(f'"""{escape_triple_quotes(tree.value)}"""')
                return
            super()._Constant(tree)
    astunparse.Unparser = ASTUnparser
    astunparse.unparser.Unparser = ASTUnparser
    from astunparse import unparse as unparse_ast
else:
    from ast import unparse as unparse_ast


"""# Package Generation
"""
import pathlib
import shutil
import nbformat
import traitlets.config
from . import exporter
DEFAULT_IGNORE_PATTERNS = ('.ipynb_checkpoints', '__pycache__', '.*')

def create_notebook_code(nb_exporter, path: pathlib.Path) -> str:
    """Return the source code for a given notebook file

    :param nb_exporter: nbconvert exporter instance
    :param path: path to notebook
    :return:
    """
    nb = nbformat.read(path, as_version=nbformat.NO_CONVERT)
    (body, resources) = nb_exporter.from_notebook_node(nb)
    return body

def build_package_component(nb_exporter, source_dir_path, dest_dir_path, ignore_patterns=None):
    """Recursively build a pure-Python package from a source tree

    :param nb_exporter: nbconvert exporter instance
    :param source_dir_path: path to current source directory
    :param dest_dir_path: path to current destination directory
    :param ignore_patterns: glob patterns of files and directories to ignore during
    recursion
    :return:
    """
    if ignore_patterns is None:
        ignore_patterns = DEFAULT_IGNORE_PATTERNS
    for path in source_dir_path.iterdir():
        if any((path.match(p) for p in ignore_patterns)):
            continue
        relative_path = path.relative_to(source_dir_path)
        if path.match('*.ipynb'):
            dest_path = dest_dir_path / relative_path.with_suffix('.py')
            source = create_notebook_code(nb_exporter, path)
            dest_path.write_text(source)
        elif path.is_dir():
            dest_path = dest_dir_path / relative_path
            dest_path.mkdir(parents=True, exist_ok=True)
            build_package_component(nb_exporter, path, dest_path)
        else:
            dest_path = dest_dir_path / relative_path
            dest_path.write_bytes(path.read_bytes())

def build_package(source_path: pathlib.Path, dest_path: pathlib.Path, ignore_patterns=None):
    """Build a pure-Python package from a literary source tree

    :param source_path: path to notebooks directory
    :param dest_path: path to generated package
    :param ignore_patterns: glob patterns of files and directories to ignore during
    recursion
    :return:
    """
    c = traitlets.config.Config()
    nb_exporter = exporter.LiteraryPythonExporter(config=c)
    if dest_path.exists():
        shutil.rmtree(dest_path)
    dest_path.mkdir(parents=True)
    build_package_component(nb_exporter, source_path, dest_path, ignore_patterns)


"""# Notebook helper namespace
"""
__all__ = ['patch', 'update_namespace']
import pathlib
import sys
from ..core import config
from . import importer
from .patch import patch
_PACKAGE_ROOT_PATH = config.find_project_path()

def _update_namespace(namespace):
    """Update a namespace with the missing metadata required to support runtime
    imports of notebooks

    :param namespace: namespace object
    :return:
    """
    namespace['__package__'] = importer.determine_package_name(pathlib.Path.cwd(), _PACKAGE_ROOT_PATH)

def load_ipython_extension(ipython):
    """Load the import hook and setup the global state for the Literary extension.
    When IPython invokes this function, the determined package root path will be
    added to `sys.path`.

    :param ipython: IPython shell instance
    """
    _update_namespace(ipython.user_ns)
    importer.install_hook(_PACKAGE_ROOT_PATH)


"""# AST Transform Routines
"""
import ast
import logging
from typing import Any, Dict, Iterable, TypeVar
import traitlets.config
logger = logging.getLogger(__name__)
T = TypeVar('T')

class IPythonTransformer(traitlets.config.Configurable):
    """Node transformer which operates upon ast.Module nodes to check for IPython
    statements.
    """
    error_if_ipython = traitlets.Bool(default_value=True).tag(config=True)

    def visit(self, node: ast.Module) -> ast.Module:
        """Transform a Python module AST node

        :param node: ast.Module object
        :return: ast.Module object
        """
        for descendant in ast.walk(node):
            if not isinstance(descendant, ast.Call):
                continue
            if not isinstance(descendant.func, ast.Name):
                continue
            if descendant.func.id == 'get_ipython':
                msg = '`get_ipython` cannot be transpiled to pure-Python. Check for magics in exported code cells'
                if self.error_if_ipython:
                    raise ValueError(msg)
                else:
                    logger.warning(msg)
        return node

class PatchTransformer(traitlets.config.Configurable):
    """Node transformer which operates upon ast.Module nodes to handle patches."""
    patch_decorator_id = traitlets.Unicode('patch').tag(config=True)

    def _is_patch_decorator(self, node: Any) -> bool:
        """Return True if the given ast node is a patch() decorator

        :param node: ast node
        :return:
        """
        return isinstance(node, ast.Call) and isinstance(node.func, ast.Name) and (node.func.id == self.patch_decorator_id)

    def _apply_patches_to_node(self, decorated: Any, classes: Dict[str, Any]):
        """Append patch functions to body of patched class.

        :param decorated: decorated function node
        :param classes: mapping of name to class node
        :return: whether node should be removed from original AST location
        """
        patch_decorators = []
        other_decorators = []
        for node in decorated.decorator_list:
            if self._is_patch_decorator(node):
                if other_decorators:
                    raise ValueError('Applying decorators to a patched function is not safe at runtime, and is hence prohibited.')
                patch_decorators.append(node)
            else:
                other_decorators.append(node)
        if not patch_decorators:
            return False
        logger.info(f'Found patch function {ast.dump(decorated)}')
        decorated.decorator_list = other_decorators
        for node in reversed(patch_decorators):
            (patched_name_node,) = node.args
            patched_node = classes[patched_name_node.id]
            patched_node.body.append(decorated)
            logger.debug(f'Patching {patched_node.name} with {ast.dump(decorated)}')
        return True

    def _transform_module_body(self, nodes: Iterable[T]) -> Iterable[T]:
        """Transform the nodes of an ast.Module to handle patches.

        :param nodes: iterable of ast nodes
        :return:
        """
        classes = {}
        for child in nodes:
            if isinstance(child, ast.ClassDef):
                classes[child.name] = child
            if hasattr(child, 'decorator_list') and self._apply_patches_to_node(child, classes):
                continue
            yield child

    def visit(self, node: ast.Module) -> ast.Module:
        """Transform a Python module AST node.

        :param node: ast.Module object
        :return: ast.Module object
        """
        node.body = [*self._transform_module_body(node.body)]
        ast.fix_missing_locations(node)
        return node


""""""
import functools
import pathlib
import traitlets.config
from jupyter_core.paths import jupyter_config_path
CONFIG_FILE_NAME = 'literary_config'
PROJECT_ROOT_MARKERS = ('pyproject.toml', 'setup.py', 'setup.cfg', '.literary-project', f'{CONFIG_FILE_NAME}.json', f'{CONFIG_FILE_NAME}.py')

def load_config(project_path: pathlib.Path) -> traitlets.config.Config:
    """Load the literary config file, or return an empty config object if missing"""
    paths = [project_path, *jupyter_config_path()]
    loaders = [traitlets.config.JSONFileConfigLoader(f'{CONFIG_FILE_NAME}.json', path=paths), traitlets.config.PyFileConfigLoader(f'{CONFIG_FILE_NAME}.py', path=paths)]
    config = traitlets.config.Config()
    for loader in loaders:
        try:
            config.update(loader.load_config())
        except traitlets.config.ConfigFileNotFound:
            pass
    return config

@functools.lru_cache()
def find_project_path(*search_paths) -> pathlib.Path:
    """Find the root path of the literary project.

    Searches for known files / directories including
    `.git/`, `.hg/`, or `pyproject.toml`.

    :param search_paths: starting search paths
    :return:
    """
    if not search_paths:
        search_paths = [pathlib.Path.cwd()]
    visited = set()
    for search_path in search_paths:
        for path in (search_path, *search_path.parents):
            if path in visited:
                break
            visited.add(path)
            for marker in PROJECT_ROOT_MARKERS:
                if (path / marker).exists():
                    return path
    raise FileNotFoundError("Couldn't find project path")


""""""
from typing import Callable, Type, TypeVar
T = TypeVar('T')

def patch(cls: Type) -> Callable[[T], T]:
    """Decorator to monkey-patch additional methods to a class.

    At import-time, this will disappear and the source code itself will be transformed
    Inside notebooks, the implementation below will be used.

    :param cls:
    :return:
    """

    def get_name(func):
        try:
            return func.__name__
        except AttributeError:
            for attr in ('fget', 'fset', 'fdel', '__func__'):
                try:
                    return getattr(func, attr).__name__
                except AttributeError:
                    continue
            raise

    def _notebook_patch_impl(func):
        setattr(cls, get_name(func), func)
        return func
    return _notebook_patch_impl


"""# Import Hook
"""
import importlib.machinery
import inspect
import linecache
import os
import pathlib
import sys
import traceback
from typing import Any, AnyStr, Callable, Tuple, TypeVar, Union
import nbformat
from ..core.config import load_config
from ..core.exporter import LiteraryPythonExporter

class NotebookLoader(importlib.machinery.SourcelessFileLoader):
    """Sourceless Jupyter Notebook loader"""

    def __init__(self, fullname: str, path: str, config):
        super().__init__(fullname, path)
        self._config = config

    def _update_linecache(self, path: str, source: str):
        linecache.cache[path] = (len(source), None, source.splitlines(keepends=True), path)

    def get_code(self, fullname: str):
        path = self.get_filename(fullname)
        body = self.get_transpiled_source(path)
        self._update_linecache(path, body)
        return compile(body, path, 'exec')

    def get_transpiled_source(self, path: str):
        nb = nbformat.read(path, as_version=nbformat.NO_CONVERT)
        exporter = LiteraryPythonExporter(config=self._config)
        (body, resources) = exporter.from_notebook_node(nb)
        return body

def determine_package_name(path: pathlib.Path, package_root_path: pathlib.Path) -> str:
    """Determine the corresponding importable name for a package directory given by
    a particular file path

    :param path: path to package
    :param package_root_path: root path containing notebook package directory
    :return:
    """
    relative_path = path.relative_to(package_root_path)
    return '.'.join(relative_path.parts)

def _get_loader_details(hook) -> tuple:
    """Return the loader_details for a given FileFinder closure

    :param hook: FileFinder closure
    :returns: loader_details tuple
    """
    try:
        namespace = inspect.getclosurevars(hook)
    except TypeError as err:
        raise ValueError from err
    try:
        return namespace.nonlocals['loader_details']
    except KeyError as err:
        raise ValueError from err

def determine_package_name(path: pathlib.Path, package_root_path: pathlib.Path) -> str:
    """Determine the corresponding importable name for a package directory given by
    a particular file path

    :param path: path to package
    :param package_root_path: root path containing notebook package directory
    :return:
    """
    relative_path = path.relative_to(package_root_path)
    return '.'.join(relative_path.parts)

def _find_file_finder(path_hooks: list) -> Tuple[int, Any]:
    """Find the FileFinder closure in a list of path hooks

    :param path_hooks: path hooks
    :returns: index of hook and the hook itself
    """
    for (i, hook) in enumerate(path_hooks):
        try:
            _get_loader_details(hook)
        except ValueError:
            continue
        return (i, hook)
    raise ValueError
T = TypeVar('T')

def _extend_file_finder(finder: T, *loader_details) -> T:
    """Extend an existing file finder with new loader details

    :param finder: existing FileFinder instance
    :param loader_details:
    :return:
    """
    return importlib.machinery.FileFinder.path_hook(*_get_loader_details(finder), *loader_details)

def _inject_notebook_loader(path_hooks: list, loader_factory: Callable[[str, str], NotebookLoader]):
    """Inject a NotebookLoader into a list of path hooks

    :param path_hooks: list of path hooks
    :param loader_factory: factory to to create NotebookLoader
    :return:
    """
    (i, finder) = _find_file_finder(path_hooks)
    new_finder = _extend_file_finder(finder, (loader_factory, ['.ipynb']))
    path_hooks[i] = new_finder
    sys.path_importer_cache.clear()

def install_hook(package_root_path: Union[AnyStr, os.PathLike], set_except_hook: bool=True):
    """Install notebook import hook

    Don't allow the user to specify a custom search path, because we also need this to
    interoperate with the default Python module importers which use sys.path

    :param package_root_path: root path containing notebook package directory
    :param set_except_hook: overwrite `sys.excepthook` to correctly display tracebacks
    inside notebooks
    :return:
    """
    config = load_config(package_root_path)
    sys.path.append(str(package_root_path))

    def create_notebook_loader(fullname, path):
        return NotebookLoader(fullname, path, config)
    _inject_notebook_loader(sys.path_hooks, create_notebook_loader)
    if set_except_hook:
        sys.excepthook = traceback.print_exception